import pygame
import ode

from util import *
import data

class Equipment:
	# Generic item properties that all equipment shares
	# TODO: Track art and sound in here
	mass = 1.0						# The mass that this equipment adds to your ship when equipped
	name = "GenericEqName"			# The title of the object
	desc = "This space for rent"	# A description of the item to give information to the player
	cost = 1						# The base price of the item
# 	type							# Optional: A string that says what type of item it is. We might want to just use theclass of the inherited item instead.
# 	techLevel						# Optional: We could have a requirement on hulls that specify the maximum tech level allowed on a ship
# 	space							# Optional: We might want to have a "bulk" factor that means items weigh little but still take up a lot of space
	sellable = True					# True if this item can be sold, false: can't be sold

	def __init__(self, owner):
		self.owner = owner
		if (self.owner != None):
			self.ship = self.owner
			self.gameObj = self.owner.gameObj
			self.worldObj = self.owner.worldObj
	
	# This event is meant to be overridden by the implementing classes
	def update(self, dt):		# Called every frame
		pass
		
class Cargo(Equipment):
	# For now, cargo is counted as equipment.
	name = "Cargo"


# The generic thruster class that other thrusters are inherited from
class Thruster(Equipment):
	engineThrust = 20		# The force of the thruster

# TODO: Add a brake equipment for CPU and others who like to be able to stop their ship on a dime.
#		if (self.BRAKE):
#			self.stop()


# Bow thrusters turn the ship
class BowThruster(Thruster):
	rotSpeed = 2.5			# The speed at which the ship decomposes -- er, rotates

	def update(self, dt):
		# TODO: Turn the ship with a torque
		if (self.ship.LEFT_TURN):
			self.ship.bodyObj.setAngularVel((0,0,self.rotSpeed))
		elif (self.ship.RIGHT_TURN):
			self.ship.bodyObj.setAngularVel((0,0,-self.rotSpeed))
		else:
			self.ship.bodyObj.setAngularVel((0,0,0))

# Main thrusters move the ship forward
class MainThruster(Thruster): 
	engineThrust = 20		# The force of the thruster

	def update(self, dt):				
		if (self.ship.UP):
			rot = self.ship.bodyObj.getRotation()
			self.ship.bodyObj.addForce((self.engineThrust * rot[0], self.engineThrust * rot[1], 0))
		
# Auxillary thrusters allow the ship to brake and strafe left/right, as well as provide a small forward boost
class AuxThruster(Thruster):
	engineThrust = 8		# The force of the thruster

	def update(self, dt):
		rot = self.ship.bodyObj.getRotation()
		if (self.ship.UP):
			self.ship.bodyObj.addForce((self.engineThrust *  rot[0], self.engineThrust *  rot[1], 0))
		if (self.ship.LEFT_STRAFE):
			self.ship.bodyObj.addForce((self.engineThrust *  rot[1], self.engineThrust * -rot[0], 0))
		if (self.ship.RIGHT_STRAFE):
			self.ship.bodyObj.addForce((self.engineThrust * -rot[1], self.engineThrust *  rot[0], 0))
		if (self.ship.DOWN):
			self.ship.bodyObj.addForce((self.engineThrust * -rot[0], self.engineThrust * -rot[1], 0))




# The generic weapon class (for cannons, turrets and autoturrets)
class Weapon(Equipment): 
	barrelType = None
	aimerType = "Fixed"	# This is a string that can be 1 of 3 values
			# "Fixed" -- guns that are mounted facing the front of the ship
			# "MouseTurret" -- guns that track with the mouse pointer
			# "AutoTurret" -- guns that automatically choose and track the nearest target
	
	def __init__(self, owner):
		Equipment.__init__(self, owner)

		if (self.barrelType != None):
			self.barrel		= self.barrelType(owner) 	# Create an instance of the barrel type that this item shoots
		else:
			self.barrel 	= None
		
	def update(self, dt):
		if (self.ship.FIRE):

			# Aim the barrel
			rotation = [1,0,0]
			if (self.aimerType == "Fixed"):
				rotation = self.ship.bodyObj.getRotation()
			else:
				print "Error: aimerType of " + aimerType + " is not yet supported!"
				
			# Fire the barrel along the specified heading
			if (self.barrel != None):
				self.barrel.Fire( rotation )

# The generic "shooter" class that is shared between turrets and fixed cannons
class Barrel:
	refireDelay			= 0.1			# The amount of time between shots (seconds)
	recoilForce			= 2				# The amount that the ship is thrust back when the gun fires. Can be 0 for recoilless weapons.
	muzzleVelocity		= 50			# How fast the projectiles are fired from the gun
	firingAngleVariance	= 10			# The amount(in degrees) that shots can be "off" when firing. Give a large angle to give the gun larger spread(but lower accuracy).

	ammoType			= None			# The class name of the ammo that is fired. Note that this is expected to be a descendant of SpaceObject

#	self.RequiresAmmo 						# (Optional) This is to indicate things like missiles, bombs, rockets, etc that you have to buy ammo for, and the ammo type must be present in the inventory for the weapon to fire, and your inventory of the item is reduced by 1 on each fire.

	def __init__( self, owner, refireDelay = 0.1, recoilForce = 100, muzzleVelocity = 50, firingAngleVariance = 10, ammoType = None, soundFile=None ):
		self.owner = owner
		if (self.owner != None):
			self.gameObj = self.owner.gameObj
			self.worldObj = self.owner.worldObj

		self.lastFireTime = gameTime()
		
		if soundFile == None:
			self.soundFile = data.loadSound("shoot-weapon.ogg")
		else:
			self.soundFile = data.loadSound(soundFile)

	def Fire(self, rotation):
		if (self.ammoType != None):
			if ( (gameTime()) - self.lastFireTime > self.refireDelay):
				self.soundFile.play(0) #0 mens "Don't loop"
				# Create a new instance of an ammo
				newBullet = self.ammoType(self.gameObj, self.owner.bodyObj.getPosition())

				# TODO: Make firingAngleVariance actually do something!!!

				# Give it a push in the right direction with the right velocity
				xvel = rotation[0] * self.muzzleVelocity + self.owner.bodyObj.getLinearVel()[0]
				yvel = rotation[1] * self.muzzleVelocity + self.owner.bodyObj.getLinearVel()[1]
				newBullet.bodyObj.setLinearVel( (xvel, yvel, 0) )

				# If there's recoil, apply the force back to the firing ship
				if (self.recoilForce > 0):
					# Apply the force in the negative direction of where we fired the weapon
					self.owner.bodyObj.addForce((self.recoilForce * -rotation[0], self.recoilForce * -rotation[1], 0))
					
				self.lastFireTime = gameTime()

