# Base includes
import random
import math

# Game engine includes
import pygame
from pygame.locals import *
import ode

import globalvars
from globalvars import *
from gameitems import *

from pygame.locals import *

from data import *
from util import *

GEOM_SHIP		= 1
GEOM_ASTEROID 	= 2
GEOM_BULLET		= 4
GEOM_RESOURCE	= 8

class SpaceObject(pygame.sprite.Sprite):
	mass 			= 10
	spriteFile 		= "test.png"
	spriteSize		= None
	spriteGroup 	= "None"
	categoryBits 	= 0		# What kind of object categories this object belongs to
	collideBits 	= 0		# The objects that this should collide with

	lifetime 		= 0		# The amount of time (in seconds) that the projectile stays active. <= 0 for infinite
	
	geomType		= ode.GeomSphere	# What kind of shape should it use for collision? Can be GeomSphere or GeomBox
	
	hitpoints		= 1
	damage			= 0
	
	geomObj			= None
	
	dead 			= False # A flag -- don't set this, only check it
	
	def __init__(self, gameObj, pos):
		# Store when the object was created so that we can put a lifetime on objects
		self.creationTime = gameTime() 
	
		self.gameObj = gameObj
		if (self.gameObj != None):
			self.worldObj = self.gameObj.worldObj
			self.spaceObj = self.gameObj.spaceObj
		
		# PHYSICS
		self.bodyObj = ode.Body(self.worldObj)
		self.bodyObj.setPosition(pos)		
		self.pos = self.bodyObj.getPosition()
		self.heading = self._heading = getEuler(self.bodyObj.getRotation())

		# GRAPHICS
		pygame.sprite.Sprite.__init__(self) #call Sprite initializer
		self.setImage(loadImage(self.spriteFile))
		
		if   (self.spriteGroup == "Asteroid"):
			self.addToSpriteGroup(self.gameObj.asteroidGroup)
		elif (self.spriteGroup == "Bullet"):
			self.addToSpriteGroup(self.gameObj.bulletGroup)
		elif (self.spriteGroup == "Ship"):
			self.addToSpriteGroup(self.gameObj.shipGroup)
		elif (self.spriteGroup == "Particles"):
			self.addToSpriteGroup(self.gameObj.particleGroup)

		# DYNAMICS
		self.setMass(self.mass)

		# COLLISIONS
		if ((self.categoryBits != 0) and (self.collideBits != 0)):
			self.geomObj = self.geomType(self.spaceObj)
			self.geomObj.setBody(self.bodyObj)
			self.geomObj.setCategoryBits(self.categoryBits)
			self.geomObj.setCollideBits(self.collideBits)
			self.geomObj.owner = self
			self.setGeomByImage()
		
		self.update(0)

	def setImage(self, image):
		self.orig_image = self.image = image
		if (self.spriteSize != None):
			self.orig_image = self.image = pygame.transform.scale(image, self.spriteSize)
			
		self.image = pygame.transform.rotate(self.orig_image, self.heading)
		self.rect = self.image.get_rect()

	def setGeomByImage(self):		
		if (self.geomObj.__class__ is ode.GeomSphere):
			self.geomObj.setRadius(self.rect.width / 2)
		# TODO: Currently, box geoms don't work -- fix them!!!
		elif (self.geomObj.__class__ is ode.GeomBox):
			self.geomObj.setLengths((self.rect[0], self.rect[1], 1))
		
	def setMass(self, mass):
		self.massObj = ode.Mass()
#		self.massObj.setSphere(mass, 1)
		self.massObj.setSphereTotal(mass / 1000000.0, self.rect.width / 4)
		self.bodyObj.setMass(self.massObj)


	def update(self, delta_t):
		self.updateGraphics(delta_t)
		
		self.updatePosition(delta_t)
		
		self.CheckDie()


	def updateGraphics(self, delta_t):
		self.heading = getEuler(self.bodyObj.getRotation())
		if (self._heading != self.heading):
			self.image = pygame.transform.rotate(self.orig_image, self.heading)
			self._heading = self.heading

		
	def updatePosition(self, delta_t):
		self.pos = self.bodyObj.getPosition()

		# This part handles centering the sprite's rectangle so that it's at the proper place in relation to the camera focus.
		cameraPos = self.gameObj.cameraFocus

		x = self.pos[0]
		y = self.pos[1]

		# Wrap objects around the screen based on the WRAP_SIZE specified in the game.
		while (x < cameraPos[0] + screenWidth / 2 - self.gameObj.WRAP_SIZE[0]):
			x += self.gameObj.WRAP_SIZE[0] * 2
		while (x > cameraPos[0] + screenWidth / 2 + self.gameObj.WRAP_SIZE[0]):
			x -= self.gameObj.WRAP_SIZE[0] * 2

		while (y < cameraPos[1] + screenHeight / 2 - self.gameObj.WRAP_SIZE[1]):
			y += self.gameObj.WRAP_SIZE[1] * 2
		while (y > cameraPos[1] + screenHeight / 2 + self.gameObj.WRAP_SIZE[1]):
			y -= self.gameObj.WRAP_SIZE[1] * 2
		
		self.bodyObj.setPosition((x, y, 0))
		self.pos = self.bodyObj.getPosition()

		
		# Limit the spinning velocities of entities
		rotVel = self.bodyObj.getAngularVel()[2]
		if (rotVel > MAX_ROTATIONAL_VELOCITY):
			self.bodyObj.setAngularVel((0,0,rotVel))
		elif (rotVel < -MAX_ROTATIONAL_VELOCITY):
			self.bodyObj.setAngularVel((0,0,-rotVel))


		self.rect = self.image.get_rect()
		self.rect.center = (self.pos[0], self.pos[1])
		self.rect = self.rect.move(-cameraPos[0], -cameraPos[1])

	# Track the sprite group so that we can remove ourselves on a delete
	def addToSpriteGroup(self, spriteGroup):
		self.spriteGroup = spriteGroup
		spriteGroup.add(self)
	
	# Delete the object from the system -- PyODE seems to want to hang onto these objects, but this seems to be as good as we can do for now.
	def delete(self):
		if (self.spriteGroup != None):
			self.spriteGroup.remove(self)
		
		# TODO: Find a better way to remove the object from PyODE
		self.bodyObj.disable()

		if (self.geomObj != None):
			self.gameObj.geomsToRemove.append(self.geomObj)
#		self.spaceObj.remove(self.geomObj) # Calling this would throw an error, because the space is locked during a collision check.
		
		
	def OnCollide(self, collisionObj, collisionPt):
		# TODO: Perhaps take damage from the opposing object's mass
		
		# Take damage from the opposing object's damage
		self.DoDamage(collisionObj.damage, collisionPt)

	def DoDamage(self, damageAmount, collisionPt = None):
		if (collisionPt == None):
			collisionPt = self.pos
		
		self.gameObj.createEffect(Damage, collisionPt)
	
		self.hitpoints -= damageAmount

	def CheckDie(self):
		if not self.dead:
			if (self.hitpoints <= 0):
				self.Die()
				
			if (self.lifetime > 0):
				if (self.lifetime + self.creationTime < gameTime()):
					self.delete()
			
			
	def Die(self):
		self.gameObj.createEffect(Explosion, self.pos)
		self.delete()
		self.dead = True


# The generic space object that is used for resources
class Resource(SpaceObject):
	mass 			= 0.001
	spriteFile 		= "test.png"
	spriteSize 		= (48,48)
	spriteGroup 	= "Bullet"
	categoryBits 	= GEOM_RESOURCE
	collideBits 	= GEOM_SHIP

	resourceType	= "None"
	
	def OnCollide(self, collisionObj, collisionPt):
		self.delete()
		
		# TODO: Give some resources here
		if (self.resourceType == "Ore"):
			globalvars.player.addOre(1)
		elif (self.resourceType == "Mineral"):
			globalvars.player.addMineral(1)
		elif (self.resourceType == "Water"):
			globalvars.player.addWater(1)
		elif (self.resourceType == "Fuel"):
			globalvars.player.addFuel(1)
			
			
class OreResource(Resource):
	resourceType	= "Ore"
	spriteFile		= "resource_ore_1.png"
class MineralResource(Resource):
	resourceType	= "Mineral"
	spriteFile		= "resource_minerals_1.png"
class WaterResource(Resource):
	resourceType	= "Water"
	spriteFile		= "resource_water_1.png"
class FuelResource(Resource):
	resourceType	= "Fuel"
	spriteFile		= "resource_fuel_1.png"



class Asteroid(SpaceObject):
	spriteFile 		= "asteroid_rock1_1.png"
	spriteGroup 	= "Asteroid"
	resourceType	= None
	
	geomType 		= ode.GeomSphere
	
	categoryBits 	= GEOM_ASTEROID
	collideBits 	= GEOM_ASTEROID | GEOM_SHIP | GEOM_BULLET

	sizeMasses 		= [1, 2, 4, 8, 16, 32, 64, 128]
	sizeDimensions 	= [(32,32),(45,45),(64,64),(90,90),(128,128),(181,181),(256,256),(356,356)]
	sizeHitpoints	= [4, 8, 16, 32, 64, 128, 256, 512]

	def __init__(self, gameObj, pos, size = 8):
		self.setSize(size)
		self.spriteFile = self.spriteImages[random.randrange(3)]
		SpaceObject.__init__(self, gameObj, pos)
		
	def setSize(self, size):
		self.size 		= size
		self.mass 		= self.sizeMasses[size-1]
		self.spriteSize	= self.sizeDimensions[size-1]
		self.hitpoints  = self.sizeHitpoints[size-1]
		
	def Die(self):
		self.delete()

		self.gameObj.createEffect(Explosion, self.pos)
		
		if (self.size > 2):
			# Spawn little chillun asteroids here
			for cnt in range(2):
				rock_x = self.pos[0] + (cnt * 2 - 1) * self.spriteSize[0] * .7
				rock_y = self.pos[1] + (cnt * 2 - 1) * self.spriteSize[1] * .7
				newRock = self.__class__(self.gameObj, (rock_x, rock_y, 0), self.size-1)
		elif (self.size == 2):
			# Otherwise, spawn some resources
			if (self.resourceType == WaterResource):
				for cnt in range(2):
					rock_x = self.pos[0] + (cnt * 2 - 1) * self.spriteSize[0] * .7
					rock_y = self.pos[1] + (cnt * 2 - 1) * self.spriteSize[1] * .7
					newResource = self.resourceType(self.gameObj, (rock_x, rock_y, 0))
			else:
				for cnt in range(2):
					rock_x = self.pos[0] + (cnt * 2 - 1) * self.spriteSize[0] * .7
					rock_y = self.pos[1] + (cnt * 2 - 1) * self.spriteSize[1] * .7
					newRock = self.__class__(self.gameObj, (rock_x, rock_y, 0), self.size-1)
				if (self.resourceType == MineralResource):
					newResource = self.resourceType(self.gameObj, self.pos)
		elif (self.size == 1):
			if (self.resourceType == OreResource):
				newResource = self.resourceType(self.gameObj, self.pos)


class RockAsteroid(Asteroid):
	spriteFile 		= "asteroid_rock1_1.png"
	spriteImages	= ["asteroid_rock2_1.png","asteroid_rock2_2.png","asteroid_rock2_3.png"]
	resourceType	= None
	sizeHitpoints	= [4, 8, 16, 32, 64, 128, 256, 512]

class MineralAsteroid(Asteroid):
	spriteFile 		= "asteroid_mineral_1.png"
	spriteImages	= ["asteroid_mineral_1.png","asteroid_mineral_2.png","asteroid_mineral_3.png"]
	resourceType	= MineralResource
	sizeHitpoints	= [4, 8, 16, 32, 64, 128, 256, 512]
	
class IceAsteroid(Asteroid):
	spriteFile 		= "asteroid_water_1.png"
	spriteImages	= ["asteroid_water_1.png","asteroid_water_2.png","asteroid_water_1.png"]
	resourceType	= WaterResource
	sizeHitpoints	= [3, 6, 12, 24, 48, 96, 192, 384]
			
class OreAsteroid(Asteroid):
	spriteFile 		= "asteroid_ore_1.png"
	spriteImages	= ["asteroid_ore_1.png","asteroid_ore_2.png","asteroid_ore_3.png"]
	resourceType	= OreResource
	sizeHitpoints	= [35, 70, 140, 280, 560, 1120, 2240, 4480]


# The generic space object that is used for ammunition
class Ammo(SpaceObject):
# Ammo is like firing any other space object, except it collides with asteroids and enemies rather than your ship. It is tracked by the physics system, and it 
	mass 			= 1		# The mass of the bullet
	spriteFile 		= "test.png"
	spriteGroup 	= "Bullet"
	categoryBits 	= GEOM_BULLET
	collideBits 	= GEOM_ASTEROID

	damage			= 10	# The amount of damage that the item does when it collides

# 	splashDamage	# The amount of explosive damage it can do to multiple objects near its collision 
# 	blastRadius		# The radius that splash damage effects 

	
	def OnCollide(self, collisionObj, collisionPt):
		# TODO: Give some particle effects here?
		self.delete()

class GenericShip(SpaceObject):
	mass 			= 5	# The mass of the ship
	spriteFile 		= "ship1_4.png"
	spriteGroup 	= "Ship"
	categoryBits 	= GEOM_SHIP
	collideBits 	= GEOM_ASTEROID | GEOM_SHIP | GEOM_RESOURCE
	spriteSize		= ( 96, 96 )

	playerControlled 	= False		# Default to an AI ship
	equipment 			= []
	
	lastHitTime		= 0
	invulTimeout	= 1	# How long the player stays invulnerable after a hit
	
	# Reset the input keys
	SHIFT = LEFT_TURN = RIGHT_TURN = LEFT_STRAFE = RIGHT_STRAFE = UP = DOWN = FIRE = False 	

	def update(self, delta_t):
		# SpaceObject.update(self, delta_t)

		# Only set these commands if we're not AI
		# TODO: Add wasd to the controls as well (with q and e to turn, a and d to strafe)
		if (self.playerControlled):
			keys = self.gameObj.keys
			
			self.SHIFT 			=  keys.get(K_LSHIFT, False) or keys.get(K_LSHIFT, False)
			self.LEFT_TURN 		= (keys.get(K_LEFT, False)  and not self.SHIFT) or (keys.get(K_a))
			self.RIGHT_TURN 	= (keys.get(K_RIGHT, False) and not self.SHIFT) or (keys.get(K_d))
			self.LEFT_STRAFE	= (keys.get(K_LEFT, False) 	and 	self.SHIFT) or (keys.get(K_q))
			self.RIGHT_STRAFE	= (keys.get(K_RIGHT, False) and 	self.SHIFT) or (keys.get(K_e))
			self.UP 			=  keys.get(K_UP, False)	or keys.get(K_w)
			self.DOWN 			=  keys.get(K_DOWN, False)	or keys.get(K_s)
			self.FIRE 			=  keys.get(K_SPACE, False)
		
		self.processAI( delta_t )
		
		self.updateEquipment( delta_t )
		
		self.updateGraphics( delta_t )

		self.updatePosition( delta_t )

		self.CheckDie()

		
	def processAI(self, delta_t):
		# This is where we set all of the self.LEFT_TURN sorts of things based on if we're AI
		if (not self.playerControlled):
			pass

	def updateEquipment(self, delta_t):
		for item in self.equipment:
			item.update(delta_t)
			
	def OnCollide(self, collisionObj, collisionPt):
		
		if (gameTime() > self.lastHitTime + self.invulTimeout):
			self.lastHitTime = gameTime()

			# Take damage from the opposing object's mass
			shipVel = self.bodyObj.getLinearVel()
			colVel = collisionObj.bodyObj.getLinearVel()
			
			velDiff = abs(shipVel[0] - colVel[0] + shipVel[1] - colVel[1]) # TODO: Use pythagoras, but this is fine for now
			
			#print "Velocity difference is" + str(velDiff)
			
			baseDamage = collisionObj.mass	# Use other object's mass as the base
			# print "Mass is" + str(baseDamage)
			
			damage = baseDamage * velDiff / 1000
			
			self.DoDamage(damage, collisionPt)
						
			
		# Take damage from the opposing object's damage (Add + 1 to do damage from asteroids)	
		self.DoDamage(collisionObj.damage, collisionPt)
			
	def Die(self):
		self.gameObj.createEffect(Explosion, self.pos)
		self.gameObj.createEffect(Explosion, self.pos)
		self.gameObj.createEffect(Explosion, self.pos)
		self.gameObj.createEffect(Explosion, self.pos)
		self.gameObj.createEffect(Explosion, self.pos)
		self.gameObj.createEffect(Explosion, self.pos)

		self.delete()
		self.dead = True
			
			
			
			

			
			
			
			
class Particle(SpaceObject):
	mass 			= 1
	spriteGroup 	= "Particles"
	lifetime 		= 1

	spriteFile 		= "fx_sm1.png"
	spriteSize		= None


class ParticleFx1(Particle):
	spriteFile 		= "fx_sm1.png"

class ParticleFx2(Particle):
	spriteFile 		= "fx_sm2.png"

class ParticleFx3(Particle):
	spriteFile 		= "fx_sm3.png"

class ParticleFx4(Particle):
	spriteFile 		= "fx_sm4.png"

class ParticleFx5(Particle):
	spriteFile 		= "fx_sm5.png"
	
class ParticleFx6(Particle):
	spriteFile 		= "fx_sm6.png"



class Effect:
	particles		= [ParticleFx1, ParticleFx2]
	lifetimes		= [1, 2]
	lifeVariance 	= [0.5, 0.5]	
	quantities		= [5, 10]
	velocities		= [100, 100]
	velVariance		= [20, 20]
	angle			= [0, 0]
	angleVariance	= [360, 360]
	
	overallAngle 	= 0 # The overall angle that the effect is rotated


class Explosion(Effect):
	particles		= [ParticleFx3, ParticleFx5, ParticleFx1]
	lifetimes		= [1, 2, 1]
	lifeVariance 	= [0.5, 0.5, 0.5]	
	quantities		= [10, 5, 5]
	velocities		= [100, 120, 100]
	velVariance		= [20, 20, 20]
	angle			= [0, 0, 0]
	angleVariance	= [360, 360, 360]

class Damage(Effect):
	particles		= [ParticleFx2, ParticleFx4]
	lifetimes		= [0.5, 0.5]
	lifeVariance 	= [0.4, 0.4]
	quantities		= [2, 2]
	velocities		= [50, 75]
	velVariance		= [20, 20]
	angle			= [0, 0]
	angleVariance	= [360, 360]

class Flame(Effect):
	particles		= [ParticleFx3]
	lifetimes		= [0.25]
	lifeVariance	= [0.24]
	quantities		= [1]
	velocities		= [30]
	velVariance		= [20]
	angle			= [0]
	angleVariance	= [45]

			
			
			